#include "public.h"
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <termios.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>
#include <sys/wait.h>
#include <sys/ioctl.h>

#define KEY_LEFT 68
#define KEY_RIGHT 67
#define KEY_DOWN 66
#define KEY_UP 65
#define MAX_STR 128
#define CMD_CLIENT_KEY 0x05
#define BUG 0

static void client_snake_exit(int);
static void client_exit_print(char *);
static void client_init(void);
static void client_key(void);
static void *client_socket_rs(void *);   /*与server的收发*/
static void client_key_rs(void);         /*与主进程key按键的收发*/
/*static void *client_gui_rs(void *);*/      /*与server的收发*/

void write_err(char *);
char getch(void);
int snake_open_record(int);
pid_t pty_fork(int *, char *, int, 
    const struct termios *, const struct winsize*);
int socket_client(char *);
int socket_send(const void *, const unsigned int);
int socket_receive(const void *, const unsigned int);
void socket_close(void);


extern sigset_t newmask;            /*屏蔽的信号*/
extern sigset_t oldmask;            /*保存的信号*/
extern pid_t  pid_id;               /*进程ID*/
extern int    fd_pipe[2];           /*传送key的管道*/
extern int    fd_error;             /*错误保存文件*/
extern struct termios save_tty;     /*终端*/
extern struct control cetske;       /*从server传来的蛇的数据*/

static int pseudofd;                       /*pseudo-terminal device*/
static char addr[MAX_STR];                 /*server addr*/
static char c_key;                         /*client key for send server*/
static pthread_t tid;                      /*线程ID*/
static struct termios new_tty;             /*伪终端时的终端*/
static struct termios null_tty;            /*NULL tty*/
static pid_t gui_id;                       /*gui进程ID*/
static pthread_t gtid;                     /*给gui进程送入数据的线程*/


static pthread_mutex_t \
key_lock = PTHREAD_MUTEX_INITIALIZER;       /*用于读client同步*/


void
client_run(char *pad)
{
	int i;

	client_init();
	if (pad == NULL) {
		printf("请输入主机地址：");
		for (i = 0; i < MAX_STR - 1; i++) {
			addr[i] = getchar();
			if (addr[i] == '\n') {
				addr[i] = 0;
				break;
			}
		}
		if (i > MAX_STR - 2)
			client_exit_print("地址输入错误\n");
		putchar('\n');
	} else {
		strncpy(addr, pad, MAX_STR);
		addr[MAX_STR - 1] = 0;
	}

	if (socket_client(addr) < 0) {
		write_err("socket client error\n");
		client_snake_exit(1);
	}
	/*
	 * start fork
	 */
	if ((gui_id = fork()) < 0) {
		write_err("client gui fork error\n");
		client_snake_exit(1);
	} if (gui_id == 0) {
		pid_id = 0;
		close(fd_pipe[1]);
		if (signal(SIGINT, SIG_IGN) == SIG_ERR) {
			printf("cleint gui signal error\n");
			client_snake_exit(1);
		}
		if (dup2(fd_pipe[0], STDIN_FILENO) < 0)
			client_exit_print("dup error\n");
		if (execl("/usr/local/ygsnake/snake-gui", "snake-gui", (char *) 0) < 0) {
			printf("Error,snake-gui file Not find\n");
			client_snake_exit(1);
		}
	}
	if ((pid_id = pty_fork(&pseudofd, NULL, 0, &new_tty, NULL)) < 0) {
		client_exit_print("fork error\n");
	} else if (pid_id == 0) {
		close(fd_pipe[0]);
		if (pthread_create(&tid, NULL, client_socket_rs, NULL) != 0) {
			write_err("client pthread_create error\n");
			client_snake_exit(1);
		}
		client_key_rs();
	} else {
		socket_close();
		close(fd_pipe[0]);
		close(fd_pipe[1]);

		client_key();
	}
}

/*
 * 初始化信号
 */
static void
client_init()
{
	struct sigaction sig_dispose;

	if (snake_open_record(1) < 0)
		client_exit_print("open err file error\n");
	if (pipe(fd_pipe) < 0)
		client_exit_print("pipe error\n");

	if (sigemptyset(&newmask) < 0)
		client_exit_print("sigemptyset error\n");
	if (sigaddset(&newmask, SIGINT) < 0)
		client_exit_print("sigaddset error\n");
	if (sigprocmask(SIG_BLOCK, &newmask, &oldmask) < 0)
		client_exit_print("sigprocmask error\n");

	if (sigemptyset(&sig_dispose.sa_mask) < 0)
		client_exit_print("sig_dispose for isgemptyset error\n");
	if (sigaddset(&sig_dispose.sa_mask, SIGCHLD) < 0)
		client_exit_print("sig_dispose add SIGCHLD error\n");
	sig_dispose.sa_handler = client_snake_exit; /*SIG_IGN;*/
	sig_dispose.sa_flags   = 0;
	if (sigaction(SIGINT, &sig_dispose, NULL) < 0)
		client_exit_print("sigaction error\n");
	if (sigaction(SIGPIPE, &sig_dispose, NULL) < 0)
		client_exit_print("sigaction error\n");

	sig_dispose.sa_handler = client_snake_exit;
	if (sigemptyset(&sig_dispose.sa_mask) < 0)
		client_exit_print("sig_dispose for sigemptyset error\n");
	if (sigaddset(&sig_dispose.sa_mask, SIGINT) < 0)
		client_exit_print("sig_dispose add SIGINT error\n");
	sig_dispose.sa_flags  |= SA_NOCLDSTOP;
	if (sigaction(SIGCHLD, &sig_dispose, NULL) < 0)
		client_exit_print("sigaction error\n");
	if (sigprocmask(SIG_SETMASK, &oldmask, NULL) < 0)
		client_exit_print("sigprocmask error\n");
	
	if (tcgetattr(STDIN_FILENO, &save_tty) < 0)
		client_exit_print("tcgetattr error\n");
	new_tty = save_tty;
	new_tty.c_lflag &= ~(ECHO | ICANON);
	new_tty.c_cc[VMIN] = 1;
	new_tty.c_cc[VTIME] = 0;
	c_key = KEY_LEFT;
}

static void
client_exit_print(char *str)
{
	puts(str);
	client_snake_exit(1);
}

static void
client_snake_exit(int value)
{
	if (pid_id != 0)
		kill(pid_id, SIGINT);
	if (gui_id != 0) {
		kill(gui_id, SIGINT);
		waitpid(gui_id, NULL, 0);
	}
	if (fd_error != 0)
		close(fd_error);
	if (memcmp(&save_tty, &null_tty, sizeof(save_tty)) != 0)
		tcsetattr(STDIN_FILENO, TCSAFLUSH, &save_tty);
	socket_close();
	exit(value);
}

static void
client_key()
{
	char c;
	if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &new_tty) < 0)
		client_exit_print("tcsetattr error\n");

	while ((c = getch()) != 0) {
		if (c == KEY_LEFT || c == KEY_RIGHT ||
		    c == KEY_DOWN || c == KEY_UP)
			write(pseudofd, &c, sizeof(char));
	}
}

#if 0
static void *
client_gui_rs(void *value)
{
	struct control rssnake;
	
	while (read(pseudofd, &rssnake, sizeof(rssnake)) != 0) {
		write_err("pseudo yes\n");
	}
	client_snake_exit(1);
	return NULL;
}
#endif

static void
client_key_rs(void)
{
	char ctmp;

	for (;;) {
		read(STDIN_FILENO, &ctmp, sizeof(ctmp));
		pthread_mutex_lock(&key_lock);
		c_key = ctmp;
		pthread_mutex_unlock(&key_lock);
	}
}

static void *
client_socket_rs(void *value)
{
	int icon;

	for (;;) {
		if (socket_receive(&cetske, sizeof(cetske)) 
		   != sizeof(cetske)) {
			write_err("client socket_receive error\n");
			client_snake_exit(1);
		}
		if (cetske.con_command == CMD_CLIENT_KEY) {
			pthread_mutex_lock(&key_lock);
			icon = socket_send(&c_key, sizeof(c_key));
			pthread_mutex_unlock(&key_lock);
			if (icon < 0) {
				write_err("client socket_send error\n");
				client_snake_exit(1);
			}
			continue;
		}
		icon = write(fd_pipe[1], &cetske, sizeof(struct control));
		if (icon != sizeof(struct control)) {
			write_err("client gui send error\n");
			client_snake_exit(1);
		}
#if 0
		if (write(STDOUT_FILENO, &cetske, sizeof(cetske)) 
		    != sizeof(cetske)) {
			write_err("client Pseudo-terminal write error\n");
			client_snake_exit(1);
		}
#endif
	}
}
