#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>

#include <errno.h>
#include <grp.h>


#define LINUX_SYSTEM 1


#if LINUX_SYSTEM
static int
pty_open(int value)
{
	return open("/dev/ptmx", value);
}

static char *
getptsname(int fdm)
{
	int number;
	static char pts_name[32];

	if (ioctl(fdm, TIOCGPTN, &number) < 0)
		return NULL;
	snprintf(pts_name, sizeof(pts_name), "/dev/pts/%d", number);

	return pts_name;
}

static int
grantpt(int fdm)
{
	char *pts_name;

	pts_name = getptsname(fdm);

	return chmod(pts_name, S_IRUSR | S_IWUSR | S_IWGRP);
}

static int
unlockpt(int fdm)
{
	int lock = 0;

	return ioctl(fdm, TIOCSPTLCK, &lock);
}
#endif


static int
ptym_open(char *pts_name, int pts_namesz)
{
	char *ptr;
	int fdm;

	fdm = pty_open(O_RDWR);
	if (fdm < 0)
		return -1;
	if (grantpt(fdm) < 0) {
		close(fdm);
		return -2;
	}
	if (unlockpt(fdm) < 0) {
		close(fdm);
		return -3;
	}
	if ((ptr = getptsname(fdm)) == NULL) {
		close(fdm);
		return -4;
	}
	strncpy(pts_name, ptr, pts_namesz);
	pts_name[pts_namesz - 1] = 0;

	return fdm;
}

static int
ptys_open(char *pts_name)
{
	int fds;

	if ((fds = open(pts_name, O_RDWR)) < 0)
		return -5;
	return fds;
}

pid_t
pty_fork(int *ptrfdm, char *ptyr_name, int ptyrsz,
    const struct termios *value_termios,
    const struct winsize *value_winsize)
{
	int fdm, fds;
	pid_t pid;
	char ptys_name[32];

	if ((fdm = ptym_open(ptys_name, sizeof(ptys_name))) < 0)
		return -1;
	if (ptyr_name != NULL) {
		strncpy(ptyr_name, ptys_name, ptyrsz);
		ptyr_name[ptyrsz - 1] = 0;
	}

	if ((pid = fork()) < 0) {
		return -1;
	} else if (pid == 0) {
		if (setsid() < 0)
			return -1;
		if ((fds = ptys_open(ptys_name)) < 0)
			return -1;
		close(fdm);
#if defined(TIOCSCTTY)
		if (ioctl(fds, TIOCSCTTY, (char *) 0) < 0)
			return -1;
#endif
		if (value_termios != NULL) {
			if (tcsetattr(fds, TCSANOW, value_termios) < 0)
				return -1;
		}
		if (value_winsize != NULL) {
			if (ioctl(fds, TIOCSWINSZ, value_winsize) < 0)
				return -1;
		}

		if (dup2(fds, STDIN_FILENO) != STDIN_FILENO)
			return -1;
		if (dup2(fds, STDOUT_FILENO) != STDOUT_FILENO)
			return -1;
		if (dup2(fds, STDERR_FILENO) != STDERR_FILENO)
			return -1;
		if (fds != STDIN_FILENO && fds != STDOUT_FILENO
		    && fds != STDERR_FILENO)
		    	close(fds);
		return 0;
	} else {
		*ptrfdm = fdm;
		return pid;
	}
}
