/*
 * License : GPLv2
 */

#include "snake.h"

#define BUG_KEY     0
#define WBUG        1
#define SOCKET_KEY  1
#define PAT_SERVER  0
#define PAT_CLIENT  1
#define PAT_ONLY    2

#define CLT_WRITE   0
#define CLT_READ    1
#define CLT_FREE    2
#define DEFAULTSPEED 100000;

#define SERVER_SEND(x)    (write(clientfifo[1], x, sizeof(struct control)))
#define SERVER_RECEIVE(x) (read(serverfifo[0], x, sizeof(struct control)))
#define CLIENT_SEND(x)    (write(serverfifo[1], x, sizeof(struct control)))
#define CLIENT_RECEIVE(x) (read(clientfifo[0], x, sizeof(struct control)))

#define CTOS (write(clientfifo[1], &cetske, sizeof(cetske)))

#if BUG_KEY
	FILE *key_log;
#endif

#ifdef PATH_MAX
	static long RECORD_FILE_MAX = PATH_MAX;
#else
	static long RECORD_FILE_MAX = 0;
#endif

struct control ske;            
struct termios save_tty;        
struct flock   val_lock;         
unsigned int   snake_speed;       
unsigned char  smap[Y_MAX][X_MAX]; 
pthread_t      pth_id;       
sigset_t       newmask;       
sigset_t       oldmask;        
pid_t          pid_id;     
int	       serr;       
int            fd_pipe[2];  
int            fd_record; 
int            fdlock;     
int            skey;        
int            curskey;      
int            run_key;       
int            filtration_lr;  
int	       filtration_ud;
int            if_over;       
int            fd_error;       
int            opt_server;      
int            opt_client;       
int            opt_speed;         
int            opt_only;           
extern char    *ptropt;            

/*client data*/
struct control cetske;        
int            cetskey;      
int            cetcurskey;  
int            clientfifo[2];  
int	       serverfifo[2]; 
int            cet_pid;      




pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;

int
main(int argc, char **argv)
{
	char *stropt = OPTION;  /*参数字符串*/
	int opt_value;

	while ((opt_value = getoption(argc, argv, stropt)) != 0) {
		switch (opt_value) {
		case 'c' :
			opt_server = 1;
			break;
		case 'j' :
			opt_client = 1;
			break;
		case 's' :
			opt_speed = 1;
			break;
		default :
			if (opt_value < 0) {
				exit_print("参数设置错误\n");
			} else {
				printf("Find not option:%c\n", opt_value);
				exit(1);
			}
			break;
		}
	}

	if (opt_server && opt_client)
		exit_print("参数 c 与参数 j 不能同时存在\n");
	if (opt_speed && opt_server)
		exit_print("参数 s 与参数 c 不能同时存在\n");
	if (opt_speed && opt_client)
		exit_print("参数 s 与参数 j 不能同时存在\n");
	if (!(opt_server || opt_client))
		opt_only = 1;
	if (opt_client)
		client_run(ptropt);
		

	if (opt_speed) {
		if (ptropt == NULL)
			exit_print("速度设置错误，没有速度值(范围1-9)。\n");
	}
/*
 * socket连接测试码
 */
#if 0
	char strre[1024] = {0};
	if (opt_server) {
		if (socket_server() < 0)
			exit_print("socket_server error\n");
		socket_send("Server\n", 8);
		socket_receive(strre, sizeof(strre));
		printf("%s", strre);
		return 0;
	} else if (opt_client) {
		if (socket_client(ptropt) < 0)
			exit_print("socket_client error\n");
		socket_receive(strre, sizeof(strre));
		printf("%s", strre);
		socket_send("Client\n", 8);
		return 0;
       	}
#endif


	snake_init();

	if (opt_server) {
		int ir;
		char strr[1024];
		if ((ir = socket_server()) < 0) {
			sprintf(strr, "error = %d\n", ir);
			write_err("server socket_server error\n");
			write_err(strr);
			snake_exit(1);
		}
	}

	if ((pid_id = fork()) == 0) {
		close(fd_pipe[1]);
		if (opt_server)
			socket_close();
		if (sigprocmask(SIG_SETMASK, &oldmask, NULL) < 0)
			exit_print("sigprocmask error\n");
		if (dup2(fd_pipe[0], STDIN_FILENO) < 0)
			exit_print("dup error\n");
		if (execl("/usr/local/ygsnake/snake-gui",
		    "snake-gui", (char *) 0) < 0) {
			exit_print("Error,snake-gui file Not find\n");
		}
	} else if (pid_id < 0) {
		exit_print("fork error\n");
	}
	close(fd_pipe[0]);

	snake_start();

	return 0;
}


/*
 * 程序初始化
 * 初始化蛇头坐标、蛇尾坐标和食物坐标
 * 初始化蛇行动方向
 * 初始化图形映射
 * 初始化随机种子
 * 初始化管道
 * 初始化信号集与信号处理
 */
void
snake_init(void)
{
	struct sigaction sig_dispose;


#if BUG_KEY
	key_log = fopen("key_log", "w+");
#endif
	srand(time(NULL));
	snake_data_init();
	
	if (pipe(fd_pipe) < 0)
		exit_print("pipe error\n");

	if (sigemptyset(&newmask) < 0)
		exit_print("sigemptyset error\n");
	if (sigaddset(&newmask, SIGINT) < 0)
		exit_print("sigaddset error\n");
	if (sigprocmask(SIG_BLOCK, &newmask, &oldmask) < 0)
		exit_print("sigprocmask error\n");

	if (sigemptyset(&sig_dispose.sa_mask) < 0)
		exit_print("sig_dispose for isgemptyset error\n");
	if (sigaddset(&sig_dispose.sa_mask, SIGCHLD) < 0)
		exit_print("sig_dispose add SIGCHLD error\n");
	sig_dispose.sa_handler = snake_exit; /*SIG_IGN;*/
	sig_dispose.sa_flags   = 0;
	if (sigaction(SIGINT, &sig_dispose, NULL) < 0)
		exit_print("sigaction error\n");
	if (sigaction(SIGPIPE, &sig_dispose, NULL) < 0)
		exit_print("sigaction error\n");

	sig_dispose.sa_handler = snake_exit;
	if (sigemptyset(&sig_dispose.sa_mask) < 0)
		exit_print("sig_dispose for sigemptyset error\n");
	if (sigaddset(&sig_dispose.sa_mask, SIGINT) < 0)
		exit_print("sig_dispose add SIGINT error\n");
	sig_dispose.sa_flags  |= SA_NOCLDSTOP;
	if (sigaction(SIGCHLD, &sig_dispose, NULL) < 0)
		exit_print("sigaction error\n");
	
	if (tcgetattr(STDIN_FILENO, &save_tty) < 0)
		exit_print("tcgetattr error\n");
}

/*
 * 数据初始化
 */
void
snake_data_init(void)
{
	if (snake_open_record(0) < 0)
		exit_print("open record file error\n");
	speed_init(ptropt);
		

	memset((void *) smap, 0, sizeof(smap));
	ske.con_shead[0] = 1;
	ske.con_shead[1] = 2;
	ske.con_slast[0] = 1;
	ske.con_slast[1] = 1;
	ske.con_steady = 0;
	ske.con_stype = 0;
	smap[1][1] = 1;
	smap[1][2] = 1;
	skey = KEY_RIGHT;
	curskey = KEY_RIGHT;
#if SOCKET_KEY
	if (opt_server) {
		cetske.con_shead[0] = Y_MAX - 2;
		cetske.con_shead[1] = X_MAX - 3;
		cetske.con_slast[0] = Y_MAX - 2;
		cetske.con_slast[1] = X_MAX - 2;
		cetske.con_steady = 0;
		cetske.con_stype = 1;
		cetske.con_command = CMD_START;
		smap[Y_MAX - 2][X_MAX - 3] = 1;
		smap[Y_MAX - 2][X_MAX - 2] = 1;
		cetskey = KEY_LEFT;
		cetcurskey = KEY_LEFT;
#if 0
		val_lock.l_start = 0;
		val_lock.l_whence = SEEK_SET;
		val_lock.l_len = 0;
#endif
		if (pipe(clientfifo) < 0)
			exit_print("pipe clientfifo error\n");
		if (pipe(serverfifo) < 0)
			exit_print("pipe serverfifo error\n");
		if ((cet_pid = fork()) < 0) {
			exit_print("fork cet_pid error\n");
		} else if (cet_pid == 0) {
			close(fd_record);
			close(serverfifo[0]);
			close(clientfifo[1]);
			snake_client_queue();
		} else {
			close(serverfifo[1]);
			close(clientfifo[0]);
			SERVER_SEND(&cetske);
		}

	}
#endif
	filtration_lr = KEY_RIGHT ^ KEY_LEFT;
	filtration_ud = KEY_UP ^ KEY_DOWN;
	ske.con_command = CMD_START;
	ske.con_slen    = 2;
	ske.con_srecord = snake_read_record();
	if (ske.con_srecord < 0)
		ske.con_srecord = 0;
	
	input_queue(ske.con_slast);   /*把蛇尾坐标送入列队*/
	input_queue(ske.con_shead);   /*把蛇头坐标送入列队*/
	create_food();
}

/*
 * 初始化游戏速度
 */
void
speed_init(const char *value)
{
	int si;

	lseek(fd_record, sizeof(unsigned int), SEEK_SET);
	if (opt_speed) {
		if (strlen(value) > 1)
			goto SPEED_ERROR;
		si = *value - '0';
		switch (si) {
		case 1 :
			snake_speed = 200000;
			break;
		case 2 :
			snake_speed = 135000;
			break;
		case 3 :
			snake_speed = 130000;
			break;
		case 4 :
			snake_speed = 125000;
			break;
		case 5 :
			snake_speed = 120000;
			break;
		case 6 :
			snake_speed = 115000;
			break;
		case 7 :
			snake_speed = 110000;
			break;
		case 8 :
			snake_speed = 105000;
			break;
		case 9 :
			snake_speed = 100000;
			break;
		default :
SPEED_ERROR:
			exit_print("游戏速度设置错误，速度范围只能是1-9.\n");
		}
		snake_speed ^= 0x55555555;
		write(fd_record, &snake_speed, sizeof(snake_speed));
		if (opt_only) {
			printf("游戏速度已经设置成功，请重启游戏\n");
			exit(0);
		}
	} else {
		read(fd_record, &snake_speed, sizeof(snake_speed));
		snake_speed ^= 0x55555555;
	}
}



/*
 * 游戏开始函数
 * 主要调用两个函数
 * snake_key()负责接收玩家操作
 * snake_alarm()用于倒计时（控制蛇的速度）
 */
void
snake_start(void)
{
	if (sigprocmask(SIG_UNBLOCK, &newmask, NULL) < 0)
		snake_exit(1);
	while (write(fd_pipe[1], &ske, sizeof(ske)) != sizeof(ske))
		;
	if (opt_server)
		client_send_gui(&ske);
#if WBUG
	write_err("server data write queue OK\n");
#endif
	if (opt_server) { 
		while (write(fd_pipe[1], &cetske, sizeof(cetske)) != 
		    sizeof(cetske))
			;
		client_send_gui(&cetske);
#if WBUG
	write_err("cleint data  write queue OK\n");
#endif
	}



	if (pthread_create(&pth_id, NULL, snake_alarm, NULL) != 0)
		exit_print("pthread_create error\n");

	snake_key();

	snake_exit(1);
}

/*
 * 接收与过滤玩家的按键、处理暂停
 */
void
snake_key(void)
{
	int t_key;
	int cmm;
	
	if (pthread_sigmask(SIG_BLOCK, &newmask, NULL) != 0)
		snake_exit(1);
	while ((t_key = getch()) != 0) {
#if BUG_KEY
		fprintf(key_log, "%d ", t_key);
#endif
		if (if_over) {
			free_queue();
			if (opt_server) {
				while (t_key != KEY_ENTER)
					t_key = getch();
				snake_exit(0);
			}
			snake_save_record();
			while (t_key != KEY_ENTER && t_key != 'r'
			    && t_key != 'R') {
				t_key = getch();
			}
			if (t_key == KEY_ENTER)
				snake_exit(0);

			snake_data_init();
			if_over = 0;
			continue;
		}
		if (t_key == ' ' || t_key == 'p') {
			pthread_mutex_lock(&mut);
			cmm = ske.con_command;
			ske.con_command = CMD_STOP;
			while (write(fd_pipe[1], &ske, sizeof(ske)) != sizeof(ske))
				;
			ske.con_command = cmm;
			while ((t_key = getch()) != ' ' && t_key != 'p')
				;
			pthread_mutex_unlock(&mut);
			continue;
		}
		if (t_key == KEY_LEFT || t_key == KEY_RIGHT ||
		    t_key == KEY_DOWN || t_key == KEY_UP) {
			pthread_mutex_lock(&mut);
			skey = t_key;
			pthread_mutex_unlock(&mut);
		}
	}

	snake_exit(1);
}

/*
 * 食物生成
 */
void
create_food(void)
{
	int x, y;
	do {
		x = rand() % X_MAX;
		y = rand() % Y_MAX;
	} while (smap[y][x] != 0);
	ske.con_sfood[0] = y;
	ske.con_sfood[1] = x;
	smap[y][x] = 2;
}

/*
 * 蛇行动的处理
 * 蛇头坐标向前移动一位
 * 并判断是否有出界或碰上自己身体
 * 判断是否有吃到食物
 */
void
snake_run(struct control *value)
{
	int x, y;
	switch (run_key) {
		case KEY_LEFT :
			--value->con_shead[1];
			break;
		case KEY_RIGHT :
			++value->con_shead[1];
			break;
		case KEY_UP :
			--value->con_shead[0];
			break;
		case KEY_DOWN :
			++value->con_shead[0];
			break;
	}

	x = value->con_shead[1];
	y = value->con_shead[0];

	if ((y < 0 || y >= Y_MAX) || (x < 0 || x >= X_MAX)) {
		if (opt_server)
			value->con_command = CMD_OVER_SOCKET;
		else
			value->con_command = CMD_OVER;
		goto Snake_over;
	}

	switch (smap[y][x]) {
		case 0 :
			value->con_command = CMD_NORMAL;
			if (!value->con_stype) {
				if (out_queue(value->con_slast) < 0) {
					write_err("server out queue error\n");
					snake_exit(1);
				}
			} else {
#if SOCKET_KEY
				if (SERVER_SEND(&cetske) 
				    != sizeof(struct control)) {
					write_err("client write queue error\n");    
					snake_exit(1);
				}
				if (SERVER_RECEIVE(&cetske) 
				    != sizeof(struct control)) {
					write_err("client queuei return data error\n");
					snake_exit(1);
				}
#endif
			}
			smap[value->con_slast[0]][value->con_slast[1]] = 0;
			break;
		case 1 :
			if (opt_server)
				value->con_command = CMD_OVER_SOCKET;
			else
				value->con_command = CMD_OVER;
			break;
		case 2 :
			value->con_command = CMD_FOOD;
			++value->con_slen;
			create_food();
#if SOCKET_KEY
			if (value->con_stype) {
				value->con_sfood[0] = ske.con_sfood[0];
				value->con_sfood[1] = ske.con_sfood[1];
				if (SERVER_SEND(&cetske) 
				    != sizeof(struct control)) {
					write_err("client write queue error\n");    
					snake_exit(1);
				}
			}
#endif
			break;
	}
	smap[y][x] = 1;

	if (!value->con_stype) {
		if (input_queue(value->con_shead) < 0) {
			write_err("server input_queue error\n");
			snake_exit(1);
		}
	}

Snake_over:
	pthread_mutex_lock(&mut);
	while (write(fd_pipe[1], value, sizeof(struct control)) != sizeof(struct control))
		;
	pthread_mutex_unlock(&mut);
	if (opt_server)
		client_send_gui(value);

	if (value->con_command == CMD_OVER || value->con_command == CMD_OVER_SOCKET) {
		snake_over();
		goto Snake_over;
	}
}

/*
 * 游戏结束
 * 等待结束信号 
 */
void
snake_over(void)
{
	if_over = 1;
	while (if_over)
		;
}

/*
 * 退出序程的一些处理
 */
void
snake_exit(int value)
{
#if BUG_KEY
	fclose(key_log);
#endif
	if (cet_pid != 0)
		kill(cet_pid, SIGINT);
	if (pid_id != 0)
		kill(pid_id, SIGINT);
	if (fdlock != 0)
		close(fdlock);
	if (fd_record != 0)
		close(fd_record);
	if (fd_error != 0)
		close(fd_error);
	if (!opt_only)
		socket_close();
	free_queue();
	tcsetattr(STDIN_FILENO, TCSANOW, &save_tty);
	exit(value);
}


/*在run之前的一个处理程序*/
void
snake_server_check(struct control *psk, int *ckey, int kkey)
{
	int fudlr;
	int check_steady = *ckey;

	fudlr = *ckey ^ kkey;
	if (fudlr == filtration_lr || fudlr == filtration_ud) {
		kkey = *ckey;
	} else {
		*ckey = kkey;
	}

	if (check_steady == kkey)
		++psk->con_steady;
	else
		psk->con_steady = 0;
	run_key = kkey;
	
}

/*
 * 计时，蛇的速度
 */
void *
snake_alarm(void *value)
{
	struct timeval time_value;
	int key;
	int tmpcmd;
	char sc;
	
	for (;;) {
		time_value.tv_sec  = 0;
		time_value.tv_usec = snake_speed;
		if (select(0, NULL, NULL, NULL, &time_value) < 0)
			continue;
		pthread_mutex_lock(&mut);
		key = skey;
		pthread_mutex_unlock(&mut);
		snake_server_check(&ske, &curskey, key);
		snake_run(&ske);
#if SOCKET_KEY
		if (opt_server) {
			/*key = KEY_LEFT;*/


			cetske.con_command = CMD_CLIENT_KEY;
			if (socket_send(&cetske, sizeof(cetske)) < 0) {
				write_err("server socket_snake error\n");
				snake_exit(1);
			}
			if (socket_receive(&sc, sizeof(sc)) < 0) {
				write_err("server socket_receive error\n");
				snake_exit(1);
			}
			key = sc;
			snake_server_check(&cetske, &cetcurskey, key);
			snake_run(&cetske);
		}	
#endif
	}
	
	return NULL;
}

/*
 * 打开记录文件
 */
int
snake_open_record(int value)
{
	unsigned int  zero;
	uid_t         userid;
	char          *namepath;
	char          *tmp;
	struct passwd *puser;
	struct stat   filemode;

	userid = getuid();
	if ((puser = getpwuid(userid)) == NULL)
		return -1;
	if (RECORD_FILE_MAX == 0) {
		if ((RECORD_FILE_MAX = pathconf("/home", _PC_PATH_MAX)) < 0)
			RECORD_FILE_MAX = 1024;
	}
	namepath = alloca(RECORD_FILE_MAX);
	if (namepath == NULL)
		return -1;
	else
		tmp = namepath;
	strcpy(tmp, "/home/");
	tmp += strlen(tmp);
	strcpy(tmp, puser->pw_name);
	tmp += strlen(tmp);
	strcpy(tmp, "/.ygsnake");
	mkdir(namepath, 0775);
	tmp += strlen(tmp);
	strcpy(tmp, "/errfile");
	fd_error = open(namepath, O_WRONLY | O_APPEND | O_CREAT, 0664);
	if (fd_error < 0)
		return -1;
	if (value)
		return 0;
	strcpy(tmp, "/record_data");
	fd_record = open(namepath, O_RDWR | O_CREAT, 0664);
	if (fd_record < 0)
		return -1;
	fstat(fd_record, &filemode);
	if (filemode.st_size == 0) {
		zero = 0;
		zero ^= 0x55555555;
		write(fd_record, &zero, sizeof(zero));
		zero = DEFAULTSPEED;
		zero ^= 0x55555555;
		write(fd_record, &zero, sizeof(zero));
	}
	return 0;
}

/*
 * 读取记录
 */
int
snake_read_record(void)
{
	unsigned int buffer;

	lseek(fd_record, 0, SEEK_SET);
	read(fd_record, &buffer, sizeof(buffer));
	buffer ^= 0x55555555;
	return buffer;
}

/*
 * 保存记录
 */
void
snake_save_record(void)
{
	unsigned int buffer;

	buffer  = ske.con_slen;
	if (!(buffer > ske.con_srecord))
		return;
	if (fd_record < 0)
		return;
	buffer ^= 0x55555555;
	lseek(fd_record, 0, SEEK_SET);
	write(fd_record, &buffer, sizeof(buffer));
	fsync(fd_record);
	close(fd_record);
}

/*
 * 写入出错文件
 */
void
write_err(char *value)
{
	int i;
	struct tm *currm;
	time_t curtime;
	char strtime[100];

	if ((curtime = time(NULL)) < 0)
		return;
	if ((currm = localtime(&curtime)) == NULL)
		return;
	sprintf(strtime, "%d-%d-%d %d:%d:%02d : ", currm->tm_year + 1900, 
	    currm->tm_mon + 1, currm->tm_mday, currm->tm_hour,
	    currm->tm_min, currm->tm_sec);
	if ((i = strlen(strtime)) > 0)
		write(fd_error, strtime, i);
	if ((i = strlen(value)) > 0)
		write(fd_error, value, i);
}

#if SOCKET_KEY
/*
 * client queue data
 */
void
snake_client_queue(void)
{
	struct control scet;

	signal(SIGCHLD, snake_exit);
Client_Loop:
	write_err("client queue OK\n");
	if (CLIENT_RECEIVE(&scet) != sizeof(struct control)) {
		write_err("client_queue read data error\n");
		snake_exit(1);
	}
	if (scet.con_command == CMD_NORMAL) {
		if (out_queue(scet.con_slast) < 0) {
			write_err("client out_queue error\n");
			exit(1);
		}
		if (CLIENT_SEND(&scet) != sizeof(struct control)) {
			write_err("client_queue write data error\n");
			exit(1);
		}
	} else if (scet.con_command == CMD_START) {
		if (input_queue(scet.con_slast) < 0) {
			write_err("clinet input queue con_slast error\n");
			exit(1);
		}
		if (input_queue(scet.con_shead) < 0) {
			write_err("clinet input queue con_shead error\n");
			exit(1);
		
		}
		goto Client_Loop;
	}
	if (input_queue(scet.con_shead) < 0) {
		write_err("client input_queue error\n");
		exit(1);
	}
goto Client_Loop;
}
#endif


void
client_send_gui(struct control *value)
{
	if (socket_send(value, sizeof(struct control)) < 0) {
		write_err("server sokcet_send error\n");
		snake_exit(1);
	}
}
