#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>

static struct sockaddr serveraddr;
static int    socketfd;
static int    socket_connect;

/*
 * function statement
 */
static int socket_print(void);
static int socket_findaddress(void);
static int socket_gethostaddr(struct sockaddr *);
static int socket_connecthost(void);

/*
 * server
 */
int
socket_server(void)
{

	socketfd = socket(AF_INET, SOCK_STREAM, 0);
	if (socketfd < 0)
		return -1;

	if (socket_findaddress() < 0)
		return -1;
	
	if (socket_print() < 0)
		return -1;
	if (listen(socketfd, 1) < 0)
		return -1;
	printf("等待连接中……\n");
	socket_connect = accept(socketfd, NULL, NULL);
	if (socket_connect < 0)
		return -1;
	printf("连接成功\n");

	return 0;
}

static int
socket_print(void)
{
	char hostaddr[INET_ADDRSTRLEN];

	if (inet_ntop(AF_INET, &serveraddr, hostaddr, INET_ADDRSTRLEN)
	    == NULL)
	    	return -1;
	printf("主机地址：\n%s\n", hostaddr);
}

static int
socket_findaddress(void)
{
	struct hostent  *phost;
	struct addrinfo *aip, *ailist;
	struct addrinfo filter;
	int             err;
	
	phost = gethostent();
	if (phost == NULL && phost->h_name == NULL)
		return -1;
	filter.ai_flags     = AI_ADDRCONFIG;
	filter.ai_family    = AF_INET;
	filter.ai_socktype  = SOCK_STREAM;
	filter.ai_protocol  = 0;
	filter.ai_addrlen   = 0;
	filter.ai_addr      = NULL;
	filter.ai_canonname = NULL;
	filter.ai_next      = NULL;
	
	err = getaddrinfo(phost->h_name, NULL, &filter, &aip);
	endhostent(); /*close host info*/

	if (err != 0) {
		printf("getaddrinfo error:\n%s\n", gai_strerror(err));
		return -1;
	}
	for (ailist = aip; ailist != NULL; aip = aip->ai_next) {
		if (ntohl(((struct sockaddr_in *) ailist->ai_addr)->sin_port)
		    > 1024) {
		    	serveraddr = *ailist->ai_addr;
			if (bind(socketfd, &serveraddr, ailist->ai_addrlen) < 0)
				return -1;
			freeaddrinfo(aip);
			return 0;
		}
	}
	return -1;
}

/*
 * client
 */
int
socket_client(void)
{
	socket_connect = socket(AF_INET, SOCK_STREAM, 0);
	if (socket_connect < 0)
		return -1;
	if (socket_gethostaddr(&serveraddr) < 0)
		return -1;
	if (socket_connecthost() < 0)
		return -1;
}

static int
socket_gethostaddr(struct sockaddr *socketp)
{
	char  hostaddr[INET_ADDRSTRLEN];
	char *pstr;
	int  record;

	fputs("请输入主机地址：\n", stdout);
	fgets(hostaddr, INET_ADDRSTRLEN, stdin);

	record = 0;
	for (pstr = hostaddr; *pstr != 0; ++pstr) {
		if (*pstr == '.')
			++record;
		if (*pstr == '\n')
			*pstr = 0;
	}
	if (record != 4) {
		fputs("地址格式有错\n", stdout);
		return -1;
	}

	return inet_pton(AF_INET, hostaddr, socketp);
}

static int
socket_connecthost(void)
{
	int i;

	for (i = 0; i < 3; ++i) {
		if (connect(socket_connect, &serveraddr, 
		    sizeof(struct sockaddr)) == 0)
		    	return 0;
		sleep(1);
	}
	printf("连接失败\n");
	return -1;
}

/*
 * send and receive
 */
int
socket_send(const void *haddr, const unsigned int size)
{
	return send(socket_connect, haddr, size, 0);
}

int
socket_receive(void *buffer, const unsigned int size)
{
	return recv(socket_connect, buffer, size, 0);
}
