#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>

#define SOCKET_BUG 0

#if SOCKET_BUG
#include <stdio.h>
#endif

static struct sockaddr serveraddr;
static int    socketfd;
static int    socket_connect;

/*
 * function statement
 */
static int socket_print(void);
static int socket_findaddress(void);
static int socket_gethostaddr(struct sockaddr *, char *);
static int socket_connecthost(void);

/*
 * server
 */
int
socket_server(void)
{

	socketfd = socket(AF_INET, SOCK_STREAM, 0);
	if (socketfd < 0)
		return -1;

	if (socket_findaddress() < 0)
		return -2;
	
	if (socket_print() < 0)
		return -3;
	if (listen(socketfd, 1) < 0)
		return -4;
	printf("等待连接中……\n");
	socket_connect = accept(socketfd, NULL, NULL);
	if (socket_connect < 0)
		return -5;
	printf("连接成功\n");
	close(socketfd);

	return 0;
}

static int
socket_print(void)
{
	char hostaddr[INET_ADDRSTRLEN];
	void *ptmp;

	ptmp = &serveraddr;
	if (inet_ntop(AF_INET, &((struct sockaddr_in *) ptmp)->sin_addr, 
	    hostaddr, INET_ADDRSTRLEN) == NULL)
	    	return -1;
	printf("主机地址：\n%s\n", hostaddr);

	return 0;
}

static int
socket_findaddress(void)
{
	struct addrinfo *aip, *ailist;
	struct addrinfo filter;
	char  htne[1024];
	int   err;
	
	if (gethostname(htne, 1024) < 0)
		return -1;
	filter.ai_flags     = 0;
	filter.ai_family    = 0;
	filter.ai_socktype  = SOCK_STREAM;
	filter.ai_protocol  = IPPROTO_TCP;
	filter.ai_addrlen   = 0;
	filter.ai_addr      = NULL;
	filter.ai_canonname = NULL;
	filter.ai_next      = NULL;
	
	err = getaddrinfo(htne, NULL, &filter, &aip);

	if (err != 0) {
		printf("getaddrinfo error:\n%s\n", gai_strerror(err));
		return -1;
	}
	for (ailist = aip; ailist != NULL; ailist = ailist->ai_next) {
		if (ailist == NULL)
			break;
		if (ailist->ai_family != AF_INET)
			continue;
		((struct sockaddr_in *) ailist->ai_addr)->sin_port
		    = htons(17777);
		serveraddr = *ailist->ai_addr;
		if (bind(socketfd, &serveraddr, ailist->ai_addrlen) < 0)
			continue;
		freeaddrinfo(aip);
		return 0;
	}
	printf("主机建立失败，请稍后重试.\n");
	return -1;
}

/*
 * client
 */
int
socket_client(char *straddr)
{
	socket_connect = socket(AF_INET, SOCK_STREAM, 0);
	if (socket_gethostaddr(&serveraddr,straddr) < 0)
		return -1;
	if (socket_connecthost() < 0)
		return -1;

	return 0;
}

static int
socket_gethostaddr(struct sockaddr *socketp, char *straddr)
{
	char  hostaddr[INET_ADDRSTRLEN];
	int  record, i;

	if (straddr == NULL) {
		fputs("请输入主机地址：\n", stdout);
		fgets(hostaddr, INET_ADDRSTRLEN, stdin);
	} else {
		for (i = 0; i < INET_ADDRSTRLEN; ++i)
			if ((hostaddr[i] = *straddr++) == 0)
				break;
	}

	record = 0;
	for (i = 0; i < INET_ADDRSTRLEN; ++i) {
		if (hostaddr[i] == '.')
			++record;
		if (hostaddr[i] == '\n') {
			hostaddr[i] = 0;
			break;
		}
		if (hostaddr[i] == 0)
			break;
	}
	if (record != 3) {
		fputs("地址格式有错\n", stdout);
		return -1;
	}

	((struct sockaddr_in *) socketp)->sin_family = AF_INET;
	return inet_pton(AF_INET, hostaddr, 
	    &((struct sockaddr_in *) socketp)->sin_addr);
}

static int
socket_connecthost(void)
{
	int i;
	void *ptmp;

	ptmp = &serveraddr;
	((struct sockaddr_in *) ptmp)->sin_port = htons(17777);
		if (socket_connect < 0)
			return -1;
	for (i = 0; i < 10; ++i) {
		if (connect(socket_connect, &serveraddr, 
		    sizeof(struct sockaddr_in)) == 0)
		    	return 0;
		close(socket_connect);
		printf("重新连接中...\n");
		sleep(1);
	}
	printf("连接失败\n");
	return -1;
}

/*
 * send and receive
 */
int
socket_send(const void *haddr, const unsigned int size)
{
	return send(socket_connect, haddr, size, 0);
}

int
socket_receive(void *buffer, const unsigned int size)
{
	return recv(socket_connect, buffer, size, 0);
}

void
socket_close(void)
{
	if (socket_connect != 0)
		close(socket_connect);
}
