#include "winsnake.h"

WINDOW   *win_frame, *win_main, *win_stop, *win_gover;
struct control snake_data;
char *error_str;
int old_y, old_x;
int wreset;
void (*win_cfl)(int);


int
main(void)
{
	wins_init();
	wins_loadwin(0);
	wins_run();

	return 0;
}

/*
 * 初始化数据
 */
void
wins_init(void)
{
	struct sigaction sig_win;
	struct sigaction old_win;
	sigset_t sig_mas;

	setlocale(LC_ALL, "");
	if (sigemptyset(&sig_mas) < 0)
		exit_print("sig_mas sigemptyset error\n");
	if (sigaddset(&sig_mas, SIGINT) < 0)
		exit_print("sig_mas sigaddset error\n");
	if (sigaddset(&sig_mas, SIGCONT) < 0)
		exit_print("sig_mas sigaddset error\n");
	if (sigaddset(&sig_mas, SIGWINCH) < 0)
		exit_print("sig_mas sigaddset error\n");
	if (sigprocmask(SIG_BLOCK, &sig_mas, NULL) < 0)
		exit_print("sig_mas sigprocmask error\n");

	if (sigemptyset(&sig_win.sa_mask) < 0)
		exit_print("sig_win sigemptyset error\n");
	sig_win.sa_flags   = 0;
	sig_win.sa_handler = wins_exit;
	if (sigaction(SIGINT, &sig_win, NULL) < 0)
		exit_print("SIGINT of sigaction error\n");
	sig_win.sa_handler = wins_loadwin;
	if (sigaction(SIGCONT, &sig_win, NULL) < 0)
		exit_print("SIGCONT of sigaction error\n");

	/*curses init*/
	if (initscr() == NULL)
		exit_print("initscr of curses error\n");

	if (sigaction(SIGWINCH, &sig_win, &old_win) < 0) {
		error_str = "SIGWINCH of sigaction error\n";
		wins_exit(1);
	}
	if (old_win.sa_handler != SIG_DFL)
		win_cfl = old_win.sa_handler;
	curs_set(FALSE);
	start_color();
	cbreak();
	/*init_color(COLOR_BLUE, 0, 700, 1000);*/
	init_pair(1, COLOR_WHITE, COLOR_BLACK);
	init_pair(2, COLOR_BLUE, COLOR_YELLOW);
	init_pair(3, COLOR_WHITE, COLOR_GREEN);
	init_pair(4, COLOR_YELLOW, COLOR_BLUE);
	init_pair(5, COLOR_YELLOW, COLOR_RED);
	init_pair(6, COLOR_BLACK, COLOR_GREEN);
	init_pair(7, COLOR_RED, COLOR_GREEN);

	win_frame = newwin(WIN_Y + 2, WIN_X + 2, 0, 0);
	win_main  = newwin(WIN_Y, WIN_X, 0, 0);
	win_stop  = newwin(1, 30, 0, 0);
	win_gover = newwin(5, 40, 0, 0);
	if (win_frame == NULL && win_main == NULL && win_stop == NULL) {
		error_str = "newwin or derwin error\n";
		wins_exit(1);
	}

	WIN_COLOR(stdscr, 1);
	WIN_COLOR(win_frame, 2);
	WIN_COLOR(win_main, 3);
	WIN_COLOR(win_stop, 1);
	WIN_COLOR(win_gover, 4);
	wattron(win_gover, A_BOLD);

	if (sigprocmask(SIG_UNBLOCK, &sig_mas, NULL) < 0) {
		error_str = "sig_mas sigprocmask error\n";
		wins_exit(1);
	}
}


/*
 * 载入窗体
 */
void
wins_loadwin(int value)
{
	if (value == SIGWINCH) {
		if (win_cfl != NULL)
			win_cfl(value);
	}
	refresh();
	getmaxyx(stdscr, old_y, old_x);
	if (old_y < WIN_Y + 4 || old_x < WIN_X + 3) {
		error_str = "出错！因为屏幕太小。\n";
		wins_exit(1);
	}

	clear();
	mvaddstr(0, (old_x - 18) / 2, "YG-贪吃蛇 ver1.3.0");
	refresh();
	mvwin(win_frame, (old_y - WIN_Y - 2) / 2, (old_x - WIN_X - 2) / 2);
	mvwin(win_main, (old_y - WIN_Y) / 2, (old_x - WIN_X) / 2);
	mvwin(win_stop, (old_y - 1) / 2, (old_x - 30) / 2);
	mvwin(win_gover, (LINES - 4) / 2, (COLS - 40) / 2);
	wattron(win_frame, A_BOLD);
	box(win_frame, '|', '-');
	touchwin(win_frame);
	wrefresh(win_frame);
	touchwin(win_main);
	wrefresh(win_main);
}

/*
 * 运行函数
 * 用一个无限循环读取控制进程发来的数据
 * 然后根据指令，绘出蛇的图形
 */
void
wins_run(void)
{
	int *oldxy;
	int svoldxy[2], cetoldxy[2]; /*client and server old xy*/
	int i;
	int vcolor;

	while ((i = read(STDIN_FILENO, &snake_data, sizeof(snake_data)))) {
		if (i != sizeof(snake_data)) {
			if (i < 0)
				continue;
			else
				wins_exit(1);
		}
		if (snake_data.con_stype)
			vcolor  = 6;
		else
			vcolor  = 3;
		wattron(win_main, COLOR_PAIR(vcolor));
		if (snake_data.con_stype)
			oldxy = cetoldxy;
		else
			oldxy = svoldxy;
		switch (snake_data.con_command) {
			case CMD_START :
				if (wreset) {
					wreset = 0;
					wclear(win_main);
					wins_loadwin(0);
				}
				wins_print(win_main, snake_data.con_shead, 'O'); 
				wins_print(win_main, snake_data.con_slast, '*');
				if (snake_data.con_stype == 0)
					wins_print(win_main, snake_data.con_sfood, '$');
				oldxy[0] = snake_data.con_shead[0];
				oldxy[1] = snake_data.con_shead[1];
				continue;
			case CMD_NORMAL :
				wins_print(win_main, snake_data.con_shead, 'O'); 
				wins_print(win_main, snake_data.con_slast, ' ');
				break;
			case CMD_FOOD :
				wins_print(win_main, snake_data.con_shead, 'O'); 
				wins_print(win_main, snake_data.con_sfood, '$');
				break;
			case CMD_STOP :
				wins_print(win_stop, "游戏已暂停， 按“空格”键继续游戏", 0);
				continue;
			case CMD_OVER_SOCKET :
			case CMD_OVER :
				wreset = 1;
				wins_print(win_gover, "游戏结束", 0);
				continue;
			default :
				wins_exit(1);
		}
		wattron(win_main, COLOR_PAIR(vcolor));
		wins_print(win_main, oldxy, '*');
		oldxy[0] = snake_data.con_shead[0];
		oldxy[1] = snake_data.con_shead[1];
	}

	wins_exit(1);
}

/*
 * 打印图形
 */
void
wins_print(WINDOW *win, const void *data, const char ch)
{

	if (win == win_main) {
		int *snake_xy;
		snake_xy = (int *) data;
		if (ch == '$')
			wattron(win_main, COLOR_PAIR(7));
		mvwaddch(win, snake_xy[0], snake_xy[1], ch);
	} else if (win == win_stop) {
		char *str;
		str = (char *) data;
		wclear(win);
		waddstr(win, str);
	} else if (win == win_gover) {
		char *str;
		str = (char *) data;
		wclear(win);
		mvwaddstr(win, 0, (40 - 8) / 2, str);
		if (snake_data.con_command == CMD_OVER_SOCKET) {
			if (snake_data.con_stype) {
				wattron(win, COLOR_PAIR(3));
				mvwprintw(win, 1, (40 - 4) / 2, "白方");
				mvwprintw(win, 2, (40 - 4) / 2, "胜利");
			} else {
				wattron(win, COLOR_PAIR(6));
				mvwprintw(win, 1, (40 - 4) / 2, "黑方");
				mvwprintw(win, 2, (40 - 4) / 2, "胜利");
			}
			wattron(win, COLOR_PAIR(5));
			mvwaddstr(win, 4, (40 - 17) / 2, "[主机按Enetr退出]");
			goto Over_Go;
		}
		mvwprintw(win, 1, 0, "最高记录：%d", snake_data.con_srecord);
		mvwprintw(win, 2, 0, "当前长度：%d", snake_data.con_slen);
		if (snake_data.con_slen > snake_data.con_srecord)
			mvwprintw(win, 3, 0, "恭喜！新记录诞生！！");
		wattron(win, COLOR_PAIR(5));
		mvwaddstr(win, 4, 8, "[按Enetr退出]");
		mvwaddstr(win, 4, 22, "[按R重试]");
		wattron(win, COLOR_PAIR(4));
	}
Over_Go:
	mvwprintw(win_frame, WIN_Y + 1, 0, "Len:%02d  Steady:%02d", 
	    snake_data.con_slen, snake_data.con_steady);
	wrefresh(win_frame);
	touchwin(win);
	wrefresh(win);
}

/*
 * 退出
 */
void
wins_exit(int value)
{

	delwin(win_main);
	delwin(win_frame);
	delwin(win_stop);
	delwin(win_gover);
	endwin();
	if (error_str != NULL)
		fputs(error_str, stdout);
	exit(value);
}
